// Physically Based shading model: Lambetrtian diffuse BRDF + Cook-Torrance microfacet specular BRDF + IBL for ambient.
// This implementation is based on "Real Shading in Unreal Engine 4" SIGGRAPH 2013 course notes by Epic Games.
// See: http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf

$input v_position, v_texcoord0, v_tangentBasis

#define EPSILON 0.00001
#define IRRADIANCE_BLUR 0.65

#include "common.sh"

SAMPLER2D(s_albedoMap, 0);
SAMPLER2D(s_normalMap, 1);
SAMPLER2D(s_combinedMap, 2);
SAMPLER2D(s_environmentMap, 5);
SAMPLER2D(s_splitSumLUT, 6); // When stage 6 texture sampler is not explicitly defined by the shader JSON definition, the BRDF split sum texture will be implicity bound to it.

uniform vec4 u_metalnessMultiplier;
uniform vec4 u_roughnessMultiplier;
uniform vec4 u_ambientOcclusionPower;
uniform vec4 u_directionalLight;
uniform vec4 u_directionalLightColor;
uniform vec4 u_directionalLightIntensity;
uniform vec4 u_environmentLevels; // This uniform will implicity contain the mipmap levels of the stage 5 texture sampler in its first dimension.

const vec3 Fdielectric = vec3(0.04);
const vec3 eyePosition = vec3(0.0, 0.0, 0.0);

// Equirectangular texture 2D coordinates from 3D normal.
vec2 equirectCoords(vec3 v) {
    vec2 uv = vec2(atan(v.z, v.x), asin(v.y));
    uv *= vec2(0.5 / M_PI, 1.0 / M_PI);
    uv += 0.5;
    return uv;
}

// GGX/Towbridge-Reitz normal distribution function.
// Uses Disney's reparametrization of alpha = roughness^2.
float ndfGGX(float cosLh, float roughness) {
	float alpha   = roughness * roughness;
	float alphaSq = alpha * alpha;
	float denom = (cosLh * cosLh) * (alphaSq - 1.0) + 1.0;
	return alphaSq / (M_PI * denom * denom);
}

// Single term for separable Schlick-GGX below.
float gaSchlickG1(float cosTheta, float k) {
	return cosTheta / (cosTheta * (1.0 - k) + k);
}

// Schlick-GGX approximation of geometric attenuation function using Smith's method.
float gaSchlickGGX(float cosLi, float cosLo, float roughness) {
	float r = roughness + 1.0;
	float k = (r * r) / 8.0; // Epic suggests using this roughness remapping for analytic lights.
	return gaSchlickG1(cosLi, k) * gaSchlickG1(cosLo, k);
}

// Shlick's approximation of the Fresnel factor.
vec3 fresnelSchlick(vec3 F0, float cosTheta) {
	return F0 + (vec3(1.0) - F0) * pow(1.0 - cosTheta, 5.0);
}

void main() {
    // Sample input textures to get shading model params.
	vec3 albedo = texture2D(s_albedoMap, v_texcoord0).rgb;
	vec3 combined = texture2D(s_combinedMap, v_texcoord0).rgb;
	float metalness = u_metalnessMultiplier.r * combined.r;
	float roughness = u_roughnessMultiplier.r * combined.g;
	float ao = pow(combined.b, u_ambientOcclusionPower.r);

    // Outgoing light direction (vector from world-space fragment position to the "eye").
	vec3 Lo = normalize(eyePosition - v_position);

    // Get current fragment's normal and transform to world space.
	vec3 N = normalize(2.0 * texture2D(s_normalMap, v_texcoord0).rgb - 1.0);
	N = normalize(v_tangentBasis * N);

    // Angle between surface normal and outgoing light direction.
	float cosLo = max(0.0, dot(N, Lo));
    
    // Specular reflection vector.
	vec3 Lr = 2.0 * cosLo * N - Lo;

    // Fresnel reflectance at normal incidence (for metals use albedo color).
	vec3 F0 = mix(Fdielectric, albedo, metalness);

    // Direct lighting calculation for analytical lights.
	vec3 directLighting = vec3(0.0);
	{
		vec3 Li = -normalize(u_directionalLight.xyz);
		vec3 Lradiance = u_directionalLightIntensity.r * u_directionalLightColor.rgb;

		// Half-vector between Li and Lo.
		vec3 Lh = normalize(Li + Lo);

		// Calculate angles between surface normal and various light vectors.
		float cosLi = max(0.0, dot(N, Li));
		float cosLh = max(0.0, dot(N, Lh));

		// Calculate Fresnel term for direct lighting. 
		vec3 F  = fresnelSchlick(F0, max(0.0, dot(Lh, Lo)));
		// Calculate normal distribution for specular BRDF.
		float D = ndfGGX(cosLh, roughness);
		// Calculate geometric attenuation for specular BRDF.
		float G = gaSchlickGGX(cosLi, cosLo, roughness);

		// Diffuse scattering happens due to light being refracted multiple times by a dielectric medium.
		// Metals on the other hand either reflect or absorb energy, so diffuse contribution is always zero.
		// To be energy conserving we must scale diffuse BRDF contribution based on Fresnel factor & metalness.
		vec3 kd = mix(vec3(1.0) - F, vec3(0.0), metalness);

		// Lambert diffuse BRDF.
		// We don't scale by 1/PI for lighting & material units to be more convenient.
		// See: https://seblagarde.wordpress.com/2012/01/08/pi-or-not-to-pi-in-game-lighting-equation/
		vec3 diffuseBRDF = kd * albedo;

		// Cook-Torrance specular microfacet BRDF.
		vec3 specularBRDF = (F * D * G) / max(EPSILON, 4.0 * cosLi * cosLo);

		// Total contribution for this light.
		directLighting += (diffuseBRDF + specularBRDF) * Lradiance * cosLi;
	}

    // Ambient lighting (IBL).
	vec3 ambientLighting;
	{
        // Get environment map texture levels.
		float environmentLevels = u_environmentLevels.r - 1.0;

		// Sample diffuse irradiance at normal direction.
		vec3 irradiance = texture2DLod(s_environmentMap, equirectCoords(N), IRRADIANCE_BLUR * environmentLevels).rgb;

		// Calculate Fresnel term for ambient lighting.
		// Since we use pre-filtered cubemap(s) and irradiance is coming from many directions
		// use cosLo instead of angle with light's half-vector (cosLh above).
		// See: https://seblagarde.wordpress.com/2011/08/17/hello-world/
		vec3 F = fresnelSchlick(F0, cosLo);

		// Get diffuse contribution factor (as with direct lighting).
		vec3 kd = mix(vec3(1.0) - F, vec3(0.0), metalness);

		// Irradiance map contains exitant radiance assuming Lambertian BRDF, no need to scale by 1/PI here either.
		vec3 diffuseIBL = kd * albedo * irradiance;

		// Sample pre-filtered specular reflection environment at correct mipmap level.
		vec3 specularIrradiance = texture2DLod(s_environmentMap, equirectCoords(Lr), roughness * environmentLevels).rgb;

		// Split-sum approximation factors for Cook-Torrance specular BRDF.
		vec2 specularBRDF = texture2D(s_splitSumLUT, vec2(cosLo, roughness)).rg;

		// Total specular IBL contribution.
		vec3 specularIBL = (F0 * specularBRDF.x + specularBRDF.y) * specularIrradiance;

		// Total ambient lighting contribution.
		ambientLighting = diffuseIBL + specularIBL;
	}

	// Final fragment color.
	vec3 color = (directLighting + ambientLighting) * ao;
	// color = toReinhard(color); // Tone mapping (Reinhard).
	color = toGamma(color); // Gamma correction.
	gl_FragColor = vec4(color, 1.0);
}
